@@ -1,15 +1,18 @@
 use crate::config::{Config, Delimiter};
 use crate::select::SelectColumns;
-use crate::util;
+use crate::CliError;
 use crate::CliResult;
+use crate::{regex_once_cell, util};
 use cached::proc_macro::cached;
 use censor::{Censor, Sex, Zealous};
-use chrono::{NaiveTime, Utc};
-use dateparser::parse_with;
+use dynfmt::Format;
 use eudex::Hash;
 use indicatif::{ProgressBar, ProgressDrawTarget};
+use log::debug;
 use once_cell::sync::OnceCell;
 use qsv_currency::Currency;
+use qsv_dateparser::parse_with_preference;
+use rayon::prelude::*;
 use regex::Regex;
 use reverse_geocoder::{Locations, ReverseGeocoder};
 use serde::Deserialize;
@@ -127,7 +130,7 @@ $ qsv apply emptyreplace --replacement Unknown Measurement file.csv
 
 DATEFMT
 Formats a recognized date column to a specified format using --formatstr. 
-See https://docs.rs/dateparser/0.1.6/dateparser/#accepted-date-formats for
+See https://github.com/jqnatividad/belt/tree/main/dateparser#accepted-date-formats for
 recognized date formats.
 See https://docs.rs/chrono/0.4.19/chrono/format/strftime/ for 
 accepted date formats for --formatstr.
@@ -146,6 +149,20 @@ Get the week number and store it in the week_number column:
 
   $ qsv apply dateformat OpenDate --formatstr '%V' --new-column week_number file.csv
 
+DYNFMT
+Dynamically constructs a new column from other columns using the --formatstr template.
+The template can contain arbitrary characters. To insert a column value, enclose the
+column name in curly braces, replacing all non-alphanumeric characters with underscores.
+
+Examples:
+Create a new column 'mailing address' from 'house number', 'street', 'city' and 'zip-code' columns:
+
+  $ qsv apply dynfmt --formatstr '{house_number} {street}, {city} {zip_code} USA' -c 'mailing address' file.csv
+
+Create a new column 'FullName' from 'FirstName', 'MI', and 'LastName' columns:
+
+  $ qsv apply dynfmt --formatstr 'Sir/Madam {FirstName} {MI}. {LastName}' -c 'FullName' file.csv
+
 GEOCODE
 Geocodes to the nearest city center point given a location column
 [i.e. a column which contains a latitude, longitude WGS84 coordinate] against
@@ -171,6 +188,7 @@ Usage:
 qsv apply operations <operations> [options] <column> [<input>]
 qsv apply emptyreplace --replacement=<string> [options] <column> [<input>]
 qsv apply datefmt [--formatstr=<string>] [options] <column> [<input>]
+qsv apply dynfmt --formatstr=<string> [options] --new-column=<name> [<input>]
 qsv apply geocode [--formatstr=<string>] [options] <column> [<input>]
 qsv apply --help
 
@@ -179,11 +197,18 @@ apply options:
     -r, --rename <name>         New name for the transformed column.
     -C, --comparand=<string>    The string to compare against for replace & similarity operations.
     -R, --replacement=<string>  The string to use for the replace & emptyreplace operations.
-    -f, --formatstr=<string>    The date format to use when formatting dates. For formats, see
-                                https://docs.rs/chrono/0.4.19/chrono/format/strftime/index.html
-                                [default: %+]
+    --prefer-dmy                Prefer to parse dates in dmy format. Otherwise, use mdy format.
+                                Only used with the DATEFMT subcommand.
+    -f, --formatstr=<string>    This option is used by several subcommands:
 
-                                the place format to use when geocoding. The available formats are:
+                                DATEFMT: The date format to use. For formats, see
+                                  https://docs.rs/chrono/0.4.19/chrono/format/strftime/index.html
+                                  [default: %+]
+
+                                DYNFMT: the template to use to construct a new column.
+
+                                GEOCODE: the place format to use with the geocode subcommand.
+                                  The available formats are:
                                   - 'city-state' (default) - e.g. Brooklyn, New York
                                   - 'city-country' - Brooklyn, US 
                                   - 'city-state-country' | 'city-admin1-country' - Brooklyn, New York US
@@ -193,6 +218,8 @@ apply options:
                                   - 'county-country' | 'admin2-country' - Kings County, US
                                   - 'county-state-country' | 'admin2-admin1-country' - Kings County, New York US
                                   - 'country' - US
+    -j, --jobs <arg>            The number of jobs to run in parallel.
+                                When not set, the number of jobs is set to the number of CPUs detected.
 
 Common options:
     -h, --help                  Display this message
@@ -204,12 +231,8 @@ Common options:
     -q, --quiet                 Don't show progress bars.
 ";
 
-macro_rules! regex {
-    ($re:literal $(,)?) => {{
-        static RE: once_cell::sync::OnceCell<regex::Regex> = once_cell::sync::OnceCell::new();
-        RE.get_or_init(|| regex::Regex::new($re).unwrap())
-    }};
-}
+// number of CSV rows to process in a batch
+const BATCH_SIZE: usize = 24_000;
 
 static OPERATIONS: &[&str] = &[
     "len",
@@ -246,13 +269,16 @@ struct Args {
     cmd_operations: bool,
     arg_operations: String,
     cmd_datefmt: bool,
+    cmd_dynfmt: bool,
     cmd_emptyreplace: bool,
     cmd_geocode: bool,
     arg_input: Option<String>,
     flag_rename: Option<String>,
     flag_comparand: String,
     flag_replacement: String,
+    flag_prefer_dmy: bool,
     flag_formatstr: String,
+    flag_jobs: Option<usize>,
     flag_new_column: Option<String>,
     flag_output: Option<String>,
     flag_no_headers: bool,
@@ -260,7 +286,6 @@ struct Args {
     flag_quiet: bool,
 }
 
-static MIDNIGHT: OnceCell<chrono::NaiveTime> = OnceCell::new();
 static CENSOR: OnceCell<Censor> = OnceCell::new();
 static LOCS: OnceCell<Locations> = OnceCell::new();
 static GEOCODER: OnceCell<ReverseGeocoder> = OnceCell::new();
@@ -268,6 +293,7 @@ static EUDEX_COMPARAND_HASH: OnceCell<eudex::Hash> = OnceCell::new();
 static REGEX_REPLACE: OnceCell<Regex> = OnceCell::new();
 static SENTIMENT_ANALYZER: OnceCell<SentimentIntensityAnalyzer> = OnceCell::new();
 
+#[inline]
 pub fn replace_column_value(
     record: &csv::StringRecord,
     column_index: usize,
@@ -314,7 +340,8 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
     let mut sim_invokes = 0_usize;
     let mut eudex_invokes = 0_usize;
     let mut sentiment_invokes = 0_usize;
-    let operations: Vec<&str> = args.arg_operations.split(',').collect();
+    let operations_lowercase = args.arg_operations.to_lowercase();
+    let operations: Vec<&str> = operations_lowercase.split(',').collect();
     if args.cmd_operations {
         for op in &operations {
             if !OPERATIONS.contains(op) {
@@ -393,67 +420,189 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
         return fail!("you can only use censor, replace, regex_replace, similarity, eudex or sentiment ONCE per operation series.");
     };
 
+    // for dynfmt, safe_headers are the "safe" version of colnames - alphanumeric only,
+    // all other chars replaced with underscore
+    // dynfmt_fields are the columns used in the dynfmt --formatstr option
+    // we prep it so we only populate the lookup vec with the index of these columns
+    // so SimpleCurlyFormat is performant
+    let mut dynfmt_fields = Vec::with_capacity(10); // 10 is a reasonable default to save allocs
+    let mut dynfmt_template = args.flag_formatstr.to_owned();
+    if args.cmd_dynfmt {
+        if args.flag_no_headers {
+            return fail!("dynfmt operation requires headers.");
+        }
+        // first, get the fields used in the dynfmt template
+        let safe_headers = util::safe_header_names(&headers, false);
+        let formatstr_re: &'static Regex = crate::regex_once_cell!(r"\{(?P<key>\w+)?\}");
+        for format_fields in formatstr_re.captures_iter(&args.flag_formatstr) {
+            dynfmt_fields.push(format_fields.name("key").unwrap().as_str());
+        }
+        // we sort the fields so we can do binary_search
+        dynfmt_fields.sort_unstable();
+        // now, get the indices of the columns for the lookup vec
+        for (i, field) in safe_headers.into_iter().enumerate() {
+            if dynfmt_fields.binary_search(&field.as_str()).is_ok() {
+                let field_with_curly = format!("{{{field}}}");
+                let field_index = format!("{{{i}}}");
+                dynfmt_template = dynfmt_template.replace(&field_with_curly, &field_index);
+            }
+        }
+        debug!("dynfmt_fields: {dynfmt_fields:?}  dynfmt_template: {dynfmt_template}");
+    }
+
+    pub enum ApplySubCmd {
+        Operations,
+        DateFmt,
+        DynFmt,
+        Geocode,
+        EmptyReplace,
+        Unknown,
+    }
+
+    let apply_cmd = if args.cmd_operations {
+        ApplySubCmd::Operations
+    } else if args.cmd_geocode {
+        ApplySubCmd::Geocode
+    } else if args.cmd_datefmt {
+        ApplySubCmd::DateFmt
+    } else if args.cmd_dynfmt {
+        ApplySubCmd::DynFmt
+    } else if args.cmd_emptyreplace {
+        ApplySubCmd::EmptyReplace
+    } else {
+        ApplySubCmd::Unknown
+    };
+
     // prep progress bar
     let progress = ProgressBar::new(0);
-    if !args.flag_quiet {
-        let record_count = util::count_rows(&rconfig);
-        util::prep_progress(&progress, record_count);
-    } else {
+    if args.flag_quiet {
         progress.set_draw_target(ProgressDrawTarget::hidden());
+    } else {
+        util::prep_progress(&progress, util::count_rows(&rconfig)?);
     }
+    let not_quiet = !args.flag_quiet;
 
+    let prefer_dmy = args.flag_prefer_dmy || rconfig.get_dmy_preference();
+
+    // amortize memory allocation by reusing record
     #[allow(unused_assignments)]
-    let mut record = csv::StringRecord::new();
-    let records_iter = rdr.records();
-    for result in records_iter {
-        record = result?;
+    let mut batch_record = csv::StringRecord::new();
 
-        if !args.flag_quiet {
-            progress.inc(1);
-        }
-        let mut cell = record[column_index].to_owned();
+    // reuse batch buffers
+    let mut batch = Vec::with_capacity(BATCH_SIZE);
+    let mut batch_results = Vec::with_capacity(BATCH_SIZE);
 
-        if args.cmd_operations {
-            apply_operations(
-                &operations,
-                &mut cell,
-                &args.flag_comparand,
-                &args.flag_replacement,
-            );
-        } else if args.cmd_emptyreplace {
-            if cell.trim().is_empty() {
-                cell = args.flag_replacement.to_string();
-            }
-        } else if args.cmd_datefmt && !cell.is_empty() {
-            let midnight = MIDNIGHT.get_or_init(|| NaiveTime::from_hms(0, 0, 0));
-            let parsed_date = parse_with(&cell, &Utc, *midnight);
-            if let Ok(format_date) = parsed_date {
-                let formatted_date = format_date.format(&args.flag_formatstr).to_string();
-                if formatted_date.ends_with("T00:00:00+00:00") {
-                    cell = formatted_date[..10].to_string();
-                } else {
-                    cell = formatted_date;
+    // set RAYON_NUM_THREADS
+    util::njobs(args.flag_jobs);
+
+    // main loop to read CSV and construct batches for parallel processing.
+    // each batch is processed via Rayon parallel iterator.
+    // loop exits when batch is empty.
+    loop {
+        for _ in 0..BATCH_SIZE {
+            match rdr.read_record(&mut batch_record) {
+                Ok(has_data) => {
+                    if has_data {
+                        batch.push(batch_record.to_owned());
+                    } else {
+                        // nothing else to add to batch
+                        break;
+                    }
+                }
+                Err(e) => {
+                    return Err(CliError::Other(format!("Error reading file: {e}")));
                 }
             }
-        } else if args.cmd_geocode && !cell.is_empty() {
-            let search_result = search_cached(&cell, &args.flag_formatstr);
-            if let Some(geocoded_result) = search_result {
-                cell = geocoded_result;
-            }
         }
 
-        match &args.flag_new_column {
-            Some(_) => {
-                record.push_field(&cell);
-            }
-            None => {
-                record = replace_column_value(&record, column_index, &cell);
-            }
+        if batch.is_empty() {
+            // break out of infinite loop when at EOF
+            break;
         }
 
-        wtr.write_record(&record)?;
-    }
-    if !args.flag_quiet {
+        // do actual apply command via Rayon parallel iterator
+        batch
+            .par_iter()
+            .map(|record_item| {
+                let mut record = record_item.clone();
+                let mut cell = record[column_index].to_owned();
+
+                match apply_cmd {
+                    ApplySubCmd::Geocode => {
+                        if !cell.is_empty() {
+                            let search_result = search_cached(&cell, &args.flag_formatstr);
+                            if let Some(geocoded_result) = search_result {
+                                cell = geocoded_result;
+                            }
+                        }
+                    }
+                    ApplySubCmd::Operations => {
+                        apply_operations(
+                            &operations,
+                            &mut cell,
+                            &args.flag_comparand,
+                            &args.flag_replacement,
+                        );
+                    }
+                    ApplySubCmd::EmptyReplace => {
+                        if cell.trim().is_empty() {
+                            cell = args.flag_replacement.to_string();
+                        }
+                    }
+                    ApplySubCmd::DateFmt => {
+                        if !cell.is_empty() {
+                            let parsed_date = parse_with_preference(&cell, prefer_dmy);
+                            if let Ok(format_date) = parsed_date {
+                                let formatted_date =
+                                    format_date.format(&args.flag_formatstr).to_string();
+                                if formatted_date.ends_with("T00:00:00+00:00") {
+                                    cell = formatted_date[..10].to_string();
+                                } else {
+                                    cell = formatted_date;
+                                }
+                            }
+                        }
+                    }
+                    ApplySubCmd::DynFmt => {
+                        if !cell.is_empty() {
+                            let mut record_vec: Vec<String> = Vec::with_capacity(record.len());
+                            for field in &record {
+                                record_vec.push(field.to_string());
+                            }
+                            if let Ok(formatted) =
+                                dynfmt::SimpleCurlyFormat.format(&dynfmt_template, record_vec)
+                            {
+                                cell = formatted.to_string();
+                            }
+                        }
+                    }
+                    ApplySubCmd::Unknown => {
+                        unreachable!("apply subcommands are always known");
+                    }
+                }
+
+                if args.flag_new_column.is_some() {
+                    record.push_field(&cell);
+                } else {
+                    record = replace_column_value(&record, column_index, &cell);
+                }
+                record
+            })
+            .collect_into_vec(&mut batch_results);
+
+        // rayon collect() guarantees original order, so we can just append results each batch
+        for result_record in &batch_results {
+            wtr.write_record(result_record)?;
+        }
+
+        if not_quiet {
+            progress.inc(batch.len() as u64);
+        }
+
+        batch.clear();
+    } // end infinite loop
+
+    if not_quiet {
         if args.cmd_geocode {
             util::update_cache_info!(progress, SEARCH_CACHED);
         }
@@ -477,7 +626,7 @@ fn apply_operations(operations: &[&str], cell: &mut String, comparand: &str, rep
                 *cell = cell.to_uppercase();
             }
             "squeeze" => {
-                let squeezer: &'static Regex = regex!(r"\s+");
+                let squeezer: &'static Regex = regex_once_cell!(r"\s+");
                 *cell = squeezer.replace_all(cell, " ").to_string();
             }
             "trim" => {
@@ -513,7 +662,7 @@ fn apply_operations(operations: &[&str], cell: &mut String, comparand: &str, rep
             "censor_check" | "censor" => {
                 let censor = CENSOR.get_or_init(|| {
                     let mut censored_words = Censor::Standard + Zealous + Sex;
-                    for (_pos, word) in comparand.split(',').enumerate() {
+                    for word in comparand.split(',') {
                         censored_words += word.trim();
                     }
                     censored_words
@@ -570,7 +719,7 @@ fn apply_operations(operations: &[&str], cell: &mut String, comparand: &str, rep
                 let sentiment_analyzer = SENTIMENT_ANALYZER
                     .get_or_init(vader_sentiment::SentimentIntensityAnalyzer::new);
                 let sentiment_scores = sentiment_analyzer.polarity_scores(cell);
-                *cell = sentiment_scores.get("compound").unwrap().to_string();
+                *cell = sentiment_scores.get("compound").unwrap_or(&0.0).to_string();
             }
             "whatlang" => {
                 let lang_info = detect(cell);
@@ -599,58 +748,52 @@ fn search_cached(cell: &str, formatstr: &str) -> Option<String> {
     let geocoder =
         GEOCODER.get_or_init(|| ReverseGeocoder::new(LOCS.get_or_init(Locations::from_memory)));
 
-    let locregex: &'static Regex =
-        regex!(r"(?-u)([+-]?[0-9]+\.?[0-9]*|\.[0-9]+),\s*([+-]?[0-9]+\.?[0-9]*|\.[0-9]+)");
+    let locregex: &'static Regex = regex_once_cell!(
+        r"(?-u)([+-]?[0-9]+\.?[0-9]*|\.[0-9]+),\s*([+-]?[0-9]+\.?[0-9]*|\.[0-9]+)"
+    );
 
     let loccaps = locregex.captures(cell);
-    if let Some(loccaps) = loccaps {
+    loccaps.and_then(|loccaps| {
         let lat = loccaps[1].to_string().parse::<f64>().unwrap_or_default();
         let long = loccaps[2].to_string().parse::<f64>().unwrap_or_default();
         if (-90.0..=90.00).contains(&lat) && (-180.0..=180.0).contains(&long) {
             let search_result = geocoder.search((lat, long));
-            if let Some(locdetails) = search_result {
-                let geocoded_result = match formatstr {
-                    "%+" | "city-state" => format!(
-                        "{name}, {admin1}",
-                        name = locdetails.record.name,
-                        admin1 = locdetails.record.admin1,
-                    ),
-                    "city-country" => format!(
-                        "{name}, {cc}",
-                        name = locdetails.record.name,
-                        cc = locdetails.record.cc
-                    ),
-                    "city-state-country" | "city-admin1-country" => format!(
-                        "{name}, {admin1} {cc}",
-                        name = locdetails.record.name,
-                        admin1 = locdetails.record.admin1,
-                        cc = locdetails.record.cc
-                    ),
-                    "city" => locdetails.record.name.to_string(),
-                    "county" | "admin2" => locdetails.record.admin2.to_string(),
-                    "state" | "admin1" => locdetails.record.admin1.to_string(),
-                    "county-country" | "admin2-country" => format!(
-                        "{admin2}, {cc}",
-                        admin2 = locdetails.record.admin2,
-                        cc = locdetails.record.cc
-                    ),
-                    "county-state-country" | "admin2-admin1-country" => format!(
-                        "{admin2}, {admin1} {cc}",
-                        admin2 = locdetails.record.admin2,
-                        admin1 = locdetails.record.admin1,
-                        cc = locdetails.record.cc
-                    ),
-                    "country" => locdetails.record.cc.to_string(),
-                    _ => locdetails.record.name.to_string(),
-                };
-                Some(geocoded_result)
-            } else {
-                None
-            }
+            search_result.map(|locdetails| match formatstr {
+                "%+" | "city-state" => format!(
+                    "{name}, {admin1}",
+                    name = locdetails.record.name,
+                    admin1 = locdetails.record.admin1,
+                ),
+                "city-country" => format!(
+                    "{name}, {cc}",
+                    name = locdetails.record.name,
+                    cc = locdetails.record.cc
+                ),
+                "city-state-country" | "city-admin1-country" => format!(
+                    "{name}, {admin1} {cc}",
+                    name = locdetails.record.name,
+                    admin1 = locdetails.record.admin1,
+                    cc = locdetails.record.cc
+                ),
+                "city" => locdetails.record.name.to_string(),
+                "county" | "admin2" => locdetails.record.admin2.to_string(),
+                "state" | "admin1" => locdetails.record.admin1.to_string(),
+                "county-country" | "admin2-country" => format!(
+                    "{admin2}, {cc}",
+                    admin2 = locdetails.record.admin2,
+                    cc = locdetails.record.cc
+                ),
+                "county-state-country" | "admin2-admin1-country" => format!(
+                    "{admin2}, {admin1} {cc}",
+                    admin2 = locdetails.record.admin2,
+                    admin1 = locdetails.record.admin1,
+                    cc = locdetails.record.cc
+                ),
+                "country" => locdetails.record.cc.to_string(),
+                _ => locdetails.record.name.to_string(),
+            })
         } else {
             None
         }
-    } else {
-        None
-    }
+    })
 }
