@@ -38,6 +38,7 @@ enum options:
                              Will default to "index".
     --constant <value>       Fill a new column with the given value.
                              Changes the default column name to "constant".
+                             To specify a null value, pass the literal "<NULL>".
     --copy <column>          Name of a column to copy.
                              Changes the default column name to "{column}_copy".
     --uuid                   When set, the column will be populated with
@@ -53,6 +54,8 @@ Common options:
                              Must be a single character. (default: ,)
 "#;
 
+const NULL_VALUE: &str = "<NULL>";
+
 #[derive(Deserialize)]
 struct Args {
     arg_input: Option<String>,
@@ -109,13 +112,17 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
 
     while rdr.read_byte_record(&mut record)? {
         if let Some(constant_value) = &args.flag_constant {
-            record.push_field(constant_value.as_bytes());
+            if constant_value == NULL_VALUE {
+                record.push_field(b"");
+            } else {
+                record.push_field(constant_value.as_bytes());
+            }
         } else if copy_operation {
             record.push_field(&record[copy_index].to_vec());
         } else if args.flag_uuid {
             let id = Uuid::new_v4();
             record.push_field(
-                id.to_hyphenated()
+                id.as_hyphenated()
                     .encode_lower(&mut Uuid::encode_buffer())
                     .as_bytes(),
             );
@@ -125,6 +132,5 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
         }
         wtr.write_byte_record(&record)?;
     }
-
     Ok(wtr.flush()?)
 }
