@@ -6,6 +6,7 @@ use crate::config::{Config, Delimiter};
 use crate::index::Indexed;
 use crate::util;
 use crate::CliResult;
+use log::debug;
 use serde::Deserialize;
 
 static USAGE: &str = "
@@ -62,30 +63,29 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
     let mut sample_size = args.arg_sample_size;
 
     let mut wtr = Config::new(&args.flag_output).writer()?;
-    let sampled = match rconfig.indexed()? {
-        Some(mut idx) => {
-            if sample_size < 1.0 {
-                sample_size *= idx.count() as f64;
-            }
-            if args.flag_seed.is_none() && do_random_access(sample_size as u64, idx.count()) {
-                rconfig.write_headers(&mut *idx, &mut wtr)?;
-                sample_random_access(&mut idx, sample_size as u64)?
-            } else {
-                let mut rdr = rconfig.reader()?;
-                rconfig.write_headers(&mut rdr, &mut wtr)?;
-                sample_reservoir(&mut rdr, sample_size as u64, args.flag_seed)?
-            }
+    let sampled = if let Some(mut idx) = rconfig.indexed()? {
+        if sample_size < 1.0 {
+            sample_size *= idx.count() as f64;
         }
-        _ => {
-            if sample_size < 1.0 {
-                return fail!("Percentage sampling requires an index.");
-            }
+        if args.flag_seed.is_none() && do_random_access(sample_size as u64, idx.count()) {
+            rconfig.write_headers(&mut *idx, &mut wtr)?;
+            sample_random_access(&mut idx, sample_size as u64)?
+        } else {
             let mut rdr = rconfig.reader()?;
             rconfig.write_headers(&mut rdr, &mut wtr)?;
             sample_reservoir(&mut rdr, sample_size as u64, args.flag_seed)?
         }
+    } else {
+        debug!("no index");
+        if sample_size < 1.0 {
+            return fail!("Percentage sampling requires an index.");
+        }
+        let mut rdr = rconfig.reader()?;
+        rconfig.write_headers(&mut rdr, &mut wtr)?;
+        sample_reservoir(&mut rdr, sample_size as u64, args.flag_seed)?
     };
-    for row in sampled.into_iter() {
+
+    for row in sampled {
         wtr.write_byte_record(&row)?;
     }
     Ok(wtr.flush()?)
@@ -99,8 +99,11 @@ where
     R: io::Read + io::Seek,
     I: io::Read + io::Seek,
 {
+    debug!("doing sample_random_access");
     let mut all_indices = (0..idx.count()).collect::<Vec<_>>();
     let mut rng = ::rand::thread_rng();
+    // this non-cryptographic shuffle is sufficient for our use case
+    // as we're optimizing for performance
     SliceRandom::shuffle(&mut *all_indices, &mut rng);
 
     let mut sampled = Vec::with_capacity(sample_size as usize);
@@ -116,6 +119,7 @@ fn sample_reservoir<R: io::Read>(
     sample_size: u64,
     seed: Option<usize>,
 ) -> CliResult<Vec<csv::ByteRecord>> {
+    debug!("doing sample_reservoir");
     // The following algorithm has been adapted from:
     // https://en.wikipedia.org/wiki/Reservoir_sampling
     let mut reservoir = Vec::with_capacity(sample_size as usize);
@@ -127,12 +131,14 @@ fn sample_reservoir<R: io::Read>(
     // Seeding rng
     let mut rng: StdRng = match seed {
         None => StdRng::from_rng(rand::thread_rng()).unwrap(),
+        // the non-cryptographic seed_from_u64 is sufficient for our use case
+        // as we're optimizing for performance
         Some(seed) => StdRng::seed_from_u64(seed as u64),
     };
 
     // Now do the sampling.
     for (i, row) in records {
-        let random = rng.gen_range(0..i + 1);
+        let random = rng.gen_range(0..=i);
         if random < sample_size as usize {
             reservoir[random] = row?;
         }
@@ -141,5 +147,7 @@ fn sample_reservoir<R: io::Read>(
 }
 
 fn do_random_access(sample_size: u64, total: u64) -> bool {
-    sample_size <= (total / 10)
+    let raflag = sample_size <= (total / 10) as u64;
+    debug!("sample_size: {sample_size}, total: {total}, raflag: {raflag}");
+    raflag
 }
