@@ -4,13 +4,15 @@ use crate::select::SelectColumns;
 use crate::util;
 use crate::CliError;
 use crate::CliResult;
+use ahash::AHashMap;
 use csv::ByteRecord;
 use grex::RegExpBuilder;
+use itertools::Itertools;
 use log::{debug, error, info, warn};
 use serde::Deserialize;
 use serde_json::{json, value::Number, Map, Value};
 use stats::Frequencies;
-use std::{collections::hash_map::HashMap, collections::HashSet, fs::File, io::Write, path::Path};
+use std::{collections::HashSet, fs::File, io::Write, path::Path};
 
 macro_rules! fail {
     ($mesg:expr) => {
@@ -18,26 +20,43 @@ macro_rules! fail {
     };
 }
 
-static USAGE: &str = "
+static USAGE: &str = r#"
 Generate JSON Schema from CSV data.
 
 This command generates reference JSON Schema (Draft 7) from CSV data, 
 including validation rules based on input data range.
 
-Running `validate` command on original input CSV with generated schema should not flag any invalid records.
+Running `validate` command on original input CSV with generated schema 
+should not flag any invalid records.
 
 Generated schema file has `.schema.json` postfix appended. For example, 
 for input `mydata.csv`, schema file would be `mydata.csv.schema.json`. 
-If piped from stdin, then schema file would be `stdin.csv.schema.json`.
+If piped from stdin, then schema file would be `stdin.csv.schema.json` and
+a `stdin.csv` file will created with stdin's contents as well.
 
+Note that `stdin.csv` will be overwritten if it already exists.
 
 Usage:
     qsv schema [options] [<input>]
 
 Schema options:
-    --enum-threshold NUM       Cardinality threshold for adding enum constraints [default: 50]
-    --strict-dates             Enforce Internet Datetime format (RFC-3339) for detected datetime columns
+    --enum-threshold NUM       Cardinality threshold for adding enum
+                               constraints [default: 50]
+    --strict-dates             Enforce Internet Datetime format (RFC-3339)
+                               for detected datetime columns
     --pattern-columns <args>   Select columns to add pattern constraints
+    --dates-whitelist <list>   The case-insensitive patterns to look for when 
+                               shortlisting fields for date inference.
+                               i.e. if the field's name has any of these patterns,
+                               it is shortlisted for date inferencing.
+                               Set to "all" to inspect ALL fields for
+                               date/datetime types.
+                               [default: date,time,due,opened,closed]
+    --prefer-dmy               Prefer to parse dates in dmy format. Otherwise, use mdy format.
+    --stdout                   Send generated JSON schema file to stdout instead.
+    -j, --jobs <arg>           The number of jobs to run in parallel.
+                               When not set, the number of jobs is set to the
+                               number of CPUs detected.
 
 Common options:
     -h, --help                 Display this message
@@ -47,39 +66,53 @@ Common options:
                                appear as the header row in the output.
     -d, --delimiter <arg>      The field delimiter for reading CSV data.
                                Must be a single character. [default: ,]
-";
+"#;
 
-#[derive(Deserialize, Debug)]
+#[derive(Deserialize, Clone)]
 struct Args {
     flag_enum_threshold: usize,
     flag_strict_dates: bool,
     flag_pattern_columns: SelectColumns,
+    flag_dates_whitelist: String,
+    flag_prefer_dmy: bool,
+    flag_stdout: bool,
+    flag_jobs: Option<usize>,
     flag_no_headers: bool,
     flag_delimiter: Option<Delimiter>,
     arg_input: Option<String>,
 }
 
+const STDIN_CSV: &str = "stdin.csv";
+
 pub fn run(argv: &[&str]) -> CliResult<()> {
-    let args: Args = util::get_args(USAGE, argv)?;
+    let preargs: Args = util::get_args(USAGE, argv)?;
+    let mut args = preargs.clone();
 
-    // dbg!(&args);
-
-    let input_path = match &args.arg_input {
-        Some(path) => path,
-        None => "stdin.csv",
-    };
-    let input_filename: &str = match &args.arg_input {
-        Some(path) => Path::new(path).file_name().unwrap().to_str().unwrap(),
-        None => "stdin.csv",
+    // if using stdin, we create a stdin.csv file as stdin is not seekable and we need to
+    // open the file multiple times to compile stats/unique values, etc.
+    let (input_path, input_filename) = if preargs.arg_input.is_none() {
+        let mut stdin_file = File::create(STDIN_CSV)?;
+        let stdin = std::io::stdin();
+        let mut stdin_handle = stdin.lock();
+        std::io::copy(&mut stdin_handle, &mut stdin_file)?;
+        args.arg_input = Some(STDIN_CSV.to_string());
+        (STDIN_CSV.to_string(), STDIN_CSV.to_string())
+    } else {
+        let filename = Path::new(args.arg_input.as_ref().unwrap())
+            .file_name()
+            .unwrap()
+            .to_string_lossy()
+            .to_string();
+        (args.arg_input.clone().unwrap(), filename)
     };
 
-    let schema_output_filename = input_path.to_owned() + ".schema.json";
-    let mut schema_output_file =
-        File::create(&schema_output_filename).expect("unable to create schema output file");
+    // we can do this directly here, since args is mutable and
+    // Config has not been created yet at this point
+    args.flag_prefer_dmy = args.flag_prefer_dmy || std::env::var("QSV_PREFER_DMY").is_ok();
 
     // build schema for each field by their inferred type, min/max value/length, and unique values
     let mut properties_map: Map<String, Value> =
-        match infer_schema_from_stats(&args, input_filename) {
+        match infer_schema_from_stats(&args, &input_filename) {
             Ok(map) => map,
             Err(e) => {
                 let msg = format!("Failed to infer schema via stats and frequency: {e}");
@@ -87,7 +120,7 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
             }
         };
 
-    // generate regex patternfor selected String columns
+    // generate regex pattern for selected String columns
     let pattern_map = generate_string_patterns(&args, &properties_map)?;
 
     // enrich properties map with pattern constraint for String fields
@@ -100,7 +133,7 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
         }
     }
 
-    // generated list of required fields
+    // generate list of required fields
     let required_fields = get_required_fields(&properties_map);
 
     // create final JSON object for output
@@ -115,14 +148,32 @@ pub fn run(argv: &[&str]) -> CliResult<()> {
 
     let schema_pretty = serde_json::to_string_pretty(&schema).expect("prettify schema json");
 
-    schema_output_file
-        .write_all(schema_pretty.as_bytes())
-        .expect("unable to write schema file");
+    if args.flag_stdout {
+        let stdout = std::io::stdout();
+        let mut handle = stdout.lock();
 
-    // flush error report; file gets closed automagically when out-of-scope
-    schema_output_file.flush().unwrap();
+        handle
+            .write_all(schema_pretty.as_bytes())
+            .expect("unable to write schema file to stdout");
 
-    println!("Schema written to {schema_output_filename}");
+        handle.flush().unwrap();
+
+        info!("Schema written to stdout");
+    } else {
+        let schema_output_filename = input_path + ".schema.json";
+        let mut schema_output_file =
+            File::create(&schema_output_filename).expect("unable to create schema output file");
+
+        schema_output_file
+            .write_all(schema_pretty.as_bytes())
+            .expect("unable to write schema file");
+
+        // flush error report; file gets closed automagically when out-of-scope
+        schema_output_file.flush().unwrap();
+
+        println!("Schema written to {schema_output_filename}");
+        info!("Schema written to {schema_output_filename}");
+    }
 
     Ok(())
 }
@@ -140,8 +191,13 @@ fn infer_schema_from_stats(args: &Args, input_filename: &str) -> CliResult<Map<S
     // invoke cmd::stats
     let (csv_fields, csv_stats, stats_col_index_map) = get_stats_records(args)?;
 
+    // amortize memory allocation
+    let mut low_cardinality_column_indices: Vec<usize> =
+        Vec::with_capacity(args.flag_enum_threshold);
+
     // build column selector arg to invoke cmd::frequency with
     let column_select_arg: String = build_low_cardinality_column_selector_arg(
+        &mut low_cardinality_column_indices,
         args.flag_enum_threshold,
         &csv_fields,
         &csv_stats,
@@ -152,7 +208,12 @@ fn infer_schema_from_stats(args: &Args, input_filename: &str) -> CliResult<Map<S
     let unique_values_map = get_unique_values(args, &column_select_arg)?;
 
     // map holds "properties" object of json schema
-    let mut properties_map: Map<String, Value> = Map::new();
+    let mut properties_map: Map<String, Value> = Map::with_capacity(csv_fields.len());
+
+    // amortize memory allocations
+    let mut field_map: Map<String, Value> = Map::with_capacity(10);
+    let mut type_list: Vec<Value> = Vec::with_capacity(4);
+    let mut enum_list: Vec<Value> = Vec::with_capacity(args.flag_enum_threshold);
 
     // generate definition for each CSV column/field and add to properties_map
     for i in 0..csv_fields.len() {
@@ -179,13 +240,13 @@ fn infer_schema_from_stats(args: &Args, input_filename: &str) -> CliResult<Map<S
         // );
 
         // map for holding field definition
-        let mut field_map: Map<String, Value> = Map::new();
+        field_map.clear();
         let desc = format!("{header_string} column from {input_filename}");
         field_map.insert("description".to_string(), Value::String(desc));
 
         // use list to hold types, since optional fields get appended a "null" type
-        let mut type_list: Vec<Value> = Vec::new();
-        let mut enum_list: Vec<Value> = Vec::new();
+        type_list.clear();
+        enum_list.clear();
 
         match col_type {
             "String" => {
@@ -285,15 +346,15 @@ fn infer_schema_from_stats(args: &Args, input_filename: &str) -> CliResult<Map<S
         }
 
         if type_list.len() > 0 {
-            field_map.insert("type".to_string(), Value::Array(type_list));
+            field_map.insert("type".to_string(), Value::Array(type_list.clone()));
         }
 
         if enum_list.len() > 0 {
-            field_map.insert("enum".to_string(), Value::Array(enum_list));
+            field_map.insert("enum".to_string(), Value::Array(enum_list.clone()));
         }
 
         // add current field definition to properties map
-        properties_map.insert(header_string, Value::Object(field_map));
+        properties_map.insert(header_string, Value::Object(field_map.clone()));
     }
 
     Ok(properties_map)
@@ -301,7 +362,7 @@ fn infer_schema_from_stats(args: &Args, input_filename: &str) -> CliResult<Map<S
 
 /// get stats records from cmd::stats
 /// returns tuple (csv_fields, csv_stats, stats_col_index_map)
-fn get_stats_records(args: &Args) -> CliResult<(ByteRecord, Vec<Stats>, HashMap<String, usize>)> {
+fn get_stats_records(args: &Args) -> CliResult<(ByteRecord, Vec<Stats>, AHashMap<String, usize>)> {
     let stats_args = crate::cmd::stats::Args {
         arg_input: args.arg_input.clone(),
         flag_select: crate::select::SelectColumns::parse("").unwrap(),
@@ -311,9 +372,10 @@ fn get_stats_records(args: &Args) -> CliResult<(ByteRecord, Vec<Stats>, HashMap<
         flag_median: false,
         flag_quartiles: false,
         flag_nulls: false,
-        flag_nullcount: true,
-        flag_dates: true,
-        flag_jobs: util::max_jobs() as isize,
+        flag_infer_dates: true,
+        flag_dates_whitelist: args.flag_dates_whitelist.to_string(),
+        flag_prefer_dmy: args.flag_prefer_dmy,
+        flag_jobs: Some(util::njobs(args.flag_jobs)),
         flag_output: None,
         flag_no_headers: args.flag_no_headers,
         flag_delimiter: args.flag_delimiter,
@@ -323,23 +385,23 @@ fn get_stats_records(args: &Args) -> CliResult<(ByteRecord, Vec<Stats>, HashMap<
         Ok(o) => match o {
             None => {
                 info!("no index, triggering sequential stats");
-                stats_args.sequential_stats()
+                stats_args.sequential_stats(&stats_args.flag_dates_whitelist)
             }
             Some(idx) => {
                 info!("has index, triggering parallel stats");
-                stats_args.parallel_stats(idx)
+                stats_args.parallel_stats(&stats_args.flag_dates_whitelist, idx)
             }
         },
         Err(e) => {
             warn!("error determining if indexed, triggering sequential stats: {e}");
-            stats_args.sequential_stats()
+            stats_args.sequential_stats(&stats_args.flag_dates_whitelist)
         }
     }?;
 
     let stats_columns = stats_args.stat_headers();
     debug!("stats columns: {stats_columns:?}");
 
-    let mut stats_col_index_map = HashMap::new();
+    let mut stats_col_index_map = AHashMap::new();
 
     for (i, col) in stats_columns.iter().enumerate() {
         if col != "field" {
@@ -353,12 +415,13 @@ fn get_stats_records(args: &Args) -> CliResult<(ByteRecord, Vec<Stats>, HashMap<
 
 /// get column selector argument string for low cardinality columns
 fn build_low_cardinality_column_selector_arg(
+    low_cardinality_column_indices: &mut Vec<usize>,
     enum_cardinality_threshold: usize,
     csv_fields: &ByteRecord,
     csv_stats: &[Stats],
-    stats_col_index_map: &HashMap<String, usize>,
+    stats_col_index_map: &AHashMap<String, usize>,
 ) -> String {
-    let mut low_cardinality_column_indices = Vec::new();
+    low_cardinality_column_indices.clear();
 
     // identify low cardinality columns
     for i in 0..csv_fields.len() {
@@ -378,9 +441,8 @@ fn build_low_cardinality_column_selector_arg(
         };
     }
 
-    debug!("low cardinality columns: {low_cardinality_column_indices:?}");
+    // debug!("low cardinality columns: {low_cardinality_column_indices:?}");
 
-    use itertools::Itertools;
     let column_select_arg: String = low_cardinality_column_indices
         .iter()
         .map(ToString::to_string)
@@ -394,7 +456,7 @@ fn build_low_cardinality_column_selector_arg(
 fn get_unique_values(
     args: &Args,
     column_select_arg: &str,
-) -> CliResult<HashMap<String, Vec<String>>> {
+) -> CliResult<AHashMap<String, Vec<String>>> {
     // prepare arg for invoking cmd::frequency
     let freq_args = crate::cmd::frequency::Args {
         arg_input: args.arg_input.clone(),
@@ -402,7 +464,7 @@ fn get_unique_values(
         flag_limit: args.flag_enum_threshold,
         flag_asc: false,
         flag_no_nulls: true,
-        flag_jobs: util::max_jobs() as isize,
+        flag_jobs: Some(util::njobs(args.flag_jobs)),
         flag_output: None,
         flag_no_headers: args.flag_no_headers,
         flag_delimiter: args.flag_delimiter,
@@ -413,17 +475,17 @@ fn get_unique_values(
         _ => freq_args.sequential_ftables(),
     }?;
 
-    let unique_values_map = construct_map_of_unique_values(headers, ftables)?;
+    let unique_values_map = construct_map_of_unique_values(&headers, &ftables)?;
 
     Ok(unique_values_map)
 }
 
 /// construct map of unique values keyed by header
 fn construct_map_of_unique_values(
-    freq_csv_fields: ByteRecord,
-    frequency_tables: Vec<Frequencies<Vec<u8>>>,
-) -> CliResult<HashMap<String, Vec<String>>> {
-    let mut unique_values_map: HashMap<String, Vec<String>> = HashMap::new();
+    freq_csv_fields: &ByteRecord,
+    frequency_tables: &[Frequencies<Vec<u8>>],
+) -> CliResult<AHashMap<String, Vec<String>>> {
+    let mut unique_values_map: AHashMap<String, Vec<String>> = AHashMap::new();
 
     // iterate through fields and gather unique values for each field
     for (i, header_byte_slice) in freq_csv_fields.iter().enumerate() {
@@ -437,7 +499,7 @@ fn construct_map_of_unique_values(
         let header_string = convert_to_string(header_byte_slice)?;
 
         // sort the values so enum list so schema can be diff'ed between runs
-        unique_values.sort();
+        unique_values.sort_unstable();
 
         debug!(
             "enum[{header_string}]: len={}, val={:?}",
@@ -453,6 +515,7 @@ fn construct_map_of_unique_values(
 }
 
 /// convert byte slice to UTF8 String
+#[inline]
 fn convert_to_string(byte_slice: &[u8]) -> CliResult<String> {
     // convert csv header to string
     let string: String = match std::str::from_utf8(byte_slice) {
@@ -471,7 +534,7 @@ fn convert_to_string(byte_slice: &[u8]) -> CliResult<String> {
 
 /// determine required fields
 fn get_required_fields(properties_map: &Map<String, Value>) -> Vec<Value> {
-    let mut fields: Vec<Value> = Vec::new();
+    let mut fields: Vec<Value> = Vec::with_capacity(properties_map.len());
 
     // for CSV, all columns in original input file are assume required
     for key in properties_map.keys() {
@@ -485,7 +548,7 @@ fn get_required_fields(properties_map: &Map<String, Value>) -> Vec<Value> {
 fn generate_string_patterns(
     args: &Args,
     properties_map: &Map<String, Value>,
-) -> CliResult<HashMap<String, String>> {
+) -> CliResult<AHashMap<String, String>> {
     // standard boiler-plate for reading CSV
 
     let rconfig = Config::new(&args.arg_input)
@@ -498,7 +561,7 @@ fn generate_string_patterns(
     let headers = rdr.byte_headers()?.clone();
     let sel = rconfig.selection(&headers)?;
 
-    let mut pattern_map: HashMap<String, String> = HashMap::new();
+    let mut pattern_map: AHashMap<String, String> = AHashMap::new();
 
     // return empty pattern map when:
     //  * no columns are selected
@@ -509,7 +572,7 @@ fn generate_string_patterns(
     }
 
     // Map each Header to its unique Set of values
-    let mut unique_values_map: HashMap<String, HashSet<String>> = HashMap::new();
+    let mut unique_values_map: AHashMap<String, HashSet<String>> = AHashMap::new();
 
     #[allow(unused_assignments)]
     let mut record = csv::ByteRecord::new();
@@ -537,6 +600,7 @@ fn generate_string_patterns(
 
     debug!("unique values for eligible pattern columns: {unique_values_map:?}");
 
+    pattern_map.reserve(unique_values_map.len());
     for (header, value_set) in unique_values_map.iter() {
         // Convert Set to Vector
         let values: Vec<&String> = Vec::from_iter(value_set);
