@@ -8,7 +8,7 @@ use std::process;
 use std::time::Instant;
 
 use docopt::Docopt;
-use log::{error, info, log_enabled, Level};
+use log::{info, log_enabled, Level};
 use serde::Deserialize;
 
 #[cfg(feature = "python")]
@@ -28,14 +28,19 @@ macro_rules! wout {
 macro_rules! werr {
     ($($arg:tt)*) => ({
         use std::io::Write;
+        use log::error;
+        error!("{}", $($arg)*);
         (writeln!(&mut ::std::io::stderr(), $($arg)*)).unwrap();
     });
 }
 
 macro_rules! fail {
-    ($e:expr) => {
-        Err(::std::convert::From::from($e))
-    };
+    ($e:expr) => {{
+        use log::error;
+        let err = ::std::convert::From::from($e);
+        error!("{err}");
+        Err(err)
+    }};
 }
 
 macro_rules! command_list {
@@ -47,9 +52,11 @@ macro_rules! command_list {
     count       Count records
     dedup       Remove redundant rows
     enum        Add a new column enumerating CSV lines
+    excel       Exports an Excel sheet to a CSV
     exclude     Excludes the records in one CSV from another
     explode     Explode rows based on some column separator
-    fetch       Create a new column or fetch values from a URL column
+    extsort     Sort arbitrarily large text file
+    fetch*      Create a new column or fetch values from a URL column/template
     fill        Fill empty values
     fixlengths  Makes all records have same length
     flatten     Show one field per line
@@ -60,7 +67,7 @@ macro_rules! command_list {
     headers     Show header names
     help        Show this usage message
     index       Create CSV index for faster access
-    input       Read CSV data with special quoting rules
+    input       Read CSVs w/ special quoting, skipping, trimming & transcoding rules
     join        Join CSV files
     jsonl       Convert newline-delimited JSON files to CSV
     lua*        Execute Lua script on CSV data
@@ -76,12 +83,13 @@ macro_rules! command_list {
     searchset   Search CSV data with a regex set
     select      Select, re-order, duplicate or drop columns
     slice       Slice records from CSV
+    sniff       Quickly sniff CSV metadata
     sort        Sort CSV data in alphabetical, numerical, reverse or random order
     split       Split CSV data into many files
     stats       Infer data types and compute descriptive statistics
     table       Align CSV data into columns
     transpose   Transpose rows/columns of CSV data
-    validate    Validate CSV data with JSON Schema
+    validate    Validate CSV data for RFC4180-compliance or with JSON Schema
 
     * optional feature
 
@@ -107,10 +115,11 @@ Options:
     -u, --update         Update qsv to the latest release from GitHub.
     -h, --help           Display this message
     <command> -h         Display the command help message
-    -v, --version        Print version info, mem allocator, max_jobs, num_cpus then exit
+    -v, --version        Print version info, mem allocator, features installed, 
+                         max_jobs, num_cpus then exit
 
-Commands:",
-    command_list!()
+* sponsored by datHere - Data Infrastructure Engineering
+"
 );
 
 #[derive(Deserialize)]
@@ -131,20 +140,17 @@ fn main() {
 
     #[cfg(feature = "python")]
     if !check_python() {
-        if log_enabled!(Level::Error) {
-            error!("Python 3.8+ required.");
-        } else {
-            werr!("Python 3.8+ required.");
-        }
+        werr!("Python 3.8+ required.");
         ::std::process::exit(1);
     }
 
     let now = Instant::now();
-
-    if log_enabled!(Level::Info) {
-        let qsv_args: String = env::args().skip(1).collect::<Vec<_>>().join(" ");
-        info!("START: {qsv_args}");
-    }
+    let qsv_args: String = if log_enabled!(Level::Info) {
+        env::args().skip(1).collect::<Vec<_>>().join(" ")
+    } else {
+        "".to_string()
+    };
+    info!("START: {qsv_args}");
 
     let args: Args = Docopt::new(USAGE)
         .and_then(|d| {
@@ -155,13 +161,16 @@ fn main() {
         .unwrap_or_else(|e| e.exit());
     if args.flag_list {
         wout!(concat!("Installed commands:", command_list!()));
+        log_end(qsv_args, &now);
         return;
     } else if args.flag_envlist {
         util::show_env_vars();
+        log_end(qsv_args, &now);
         return;
     }
     if args.flag_update {
-        util::qsv_check_for_update("qsv");
+        util::qsv_check_for_update();
+        log_end(qsv_args, &now);
         return;
     }
     match args.arg_command {
@@ -172,69 +181,78 @@ fn main() {
 Please choose one of the following commands:",
                 command_list!()
             ));
-            util::qsv_check_for_update("qsv");
+            util::qsv_check_for_update();
+            log_end(qsv_args, &now);
             ::std::process::exit(0);
         }
         Some(cmd) => match cmd.run() {
             Ok(()) => {
-                if log_enabled!(Level::Info) {
-                    info!("END elapsed: {}", now.elapsed().as_secs_f32());
-                }
+                log_end(qsv_args, &now);
                 process::exit(0);
             }
             Err(CliError::Flag(err)) => err.exit(),
             Err(CliError::Csv(err)) => {
-                if log_enabled!(Level::Error) {
-                    error!("{err}");
-                } else {
-                    werr!("{err}");
-                }
+                werr!("{err}");
+                log_end(qsv_args, &now);
                 process::exit(1);
             }
             Err(CliError::Io(ref err)) if err.kind() == io::ErrorKind::BrokenPipe => {
+                log_end(qsv_args, &now);
                 process::exit(0);
             }
             Err(CliError::Io(err)) => {
-                if log_enabled!(Level::Error) {
-                    error!("{err}");
-                } else {
-                    werr!("{err}");
-                }
+                werr!("{err}");
+                log_end(qsv_args, &now);
                 process::exit(1);
             }
             Err(CliError::Other(msg)) => {
-                if log_enabled!(Level::Error) {
-                    error!("{msg}");
-                } else {
-                    werr!("{msg}");
-                }
+                werr!("{msg}");
+                log_end(qsv_args, &now);
                 process::exit(1);
             }
         },
     }
 }
 
+fn log_end(mut qsv_args: String, now: &Instant) {
+    if log_enabled!(Level::Info) {
+        let ellipsis = if qsv_args.len() > 24 {
+            qsv_args.truncate(24);
+            "..."
+        } else {
+            ""
+        };
+        info!(
+            "END \"{qsv_args}{ellipsis}\" elapsed: {}",
+            now.elapsed().as_secs_f32()
+        );
+    }
+}
+
 #[derive(Debug, Deserialize)]
 #[serde(rename_all = "lowercase")]
 enum Command {
-    #[cfg(feature = "apply")]
+    #[cfg(all(feature = "apply", not(feature = "lite")))]
     Apply,
     Behead,
     Cat,
     Count,
     Dedup,
     Enum,
+    Excel,
     Exclude,
     Explode,
+    ExtSort,
+    #[cfg(all(feature = "fetch", not(feature = "lite")))]
     Fetch,
     Fill,
     FixLengths,
     Flatten,
     Fmt,
-    #[cfg(feature = "foreach")]
+    #[cfg(all(feature = "foreach", target_family = "unix", not(feature = "lite")))]
     ForEach,
     Frequency,
-    #[cfg(feature = "generate")]
+    #[cfg(all(feature = "generate", not(feature = "lite")))]
     Generate,
     Headers,
     Help,
@@ -242,11 +260,11 @@ enum Command {
     Input,
     Join,
     Jsonl,
-    #[cfg(feature = "lua")]
+    #[cfg(all(feature = "lua", not(feature = "lite")))]
     Lua,
     Partition,
     Pseudo,
-    #[cfg(feature = "python")]
+    #[cfg(all(feature = "python", not(feature = "lite")))]
     Py,
     Rename,
     Replace,
@@ -257,6 +275,7 @@ enum Command {
     SearchSet,
     Select,
     Slice,
+    Sniff,
     Sort,
     Split,
     Stats,
@@ -279,37 +298,42 @@ impl Command {
         }
         match self {
             Command::Behead => cmd::behead::run(argv),
-            #[cfg(feature = "apply")]
+            #[cfg(all(feature = "apply", not(feature = "lite")))]
             Command::Apply => cmd::apply::run(argv),
             Command::Cat => cmd::cat::run(argv),
             Command::Count => cmd::count::run(argv),
             Command::Dedup => cmd::dedup::run(argv),
             Command::Enum => cmd::enumerate::run(argv),
+            Command::Excel => cmd::excel::run(argv),
             Command::Exclude => cmd::exclude::run(argv),
             Command::Explode => cmd::explode::run(argv),
+            Command::ExtSort => cmd::extsort::run(argv),
+            #[cfg(all(feature = "fetch", not(feature = "lite")))]
             Command::Fetch => cmd::fetch::run(argv),
+            #[cfg(all(feature = "foreach", target_family = "unix", not(feature = "lite")))]
+            Command::ForEach => cmd::foreach::run(argv),
             Command::Fill => cmd::fill::run(argv),
             Command::FixLengths => cmd::fixlengths::run(argv),
             Command::Flatten => cmd::flatten::run(argv),
             Command::Fmt => cmd::fmt::run(argv),
             Command::Frequency => cmd::frequency::run(argv),
-            #[cfg(feature = "generate")]
+            #[cfg(all(feature = "generate", not(feature = "lite")))]
             Command::Generate => cmd::generate::run(argv),
             Command::Headers => cmd::headers::run(argv),
             Command::Help => {
                 wout!("{USAGE}");
-                util::qsv_check_for_update("qsv");
+                util::qsv_check_for_update();
                 Ok(())
             }
             Command::Index => cmd::index::run(argv),
             Command::Input => cmd::input::run(argv),
             Command::Join => cmd::join::run(argv),
             Command::Jsonl => cmd::jsonl::run(argv),
-            #[cfg(feature = "lua")]
+            #[cfg(all(feature = "lua", not(feature = "lite")))]
             Command::Lua => cmd::lua::run(argv),
             Command::Partition => cmd::partition::run(argv),
             Command::Pseudo => cmd::pseudo::run(argv),
-            #[cfg(feature = "python")]
+            #[cfg(all(feature = "python", not(feature = "lite")))]
             Command::Py => cmd::python::run(argv),
             Command::Rename => cmd::rename::run(argv),
             Command::Replace => cmd::replace::run(argv),
@@ -320,14 +344,13 @@ impl Command {
             Command::SearchSet => cmd::searchset::run(argv),
             Command::Select => cmd::select::run(argv),
             Command::Slice => cmd::slice::run(argv),
+            Command::Sniff => cmd::sniff::run(argv),
             Command::Sort => cmd::sort::run(argv),
             Command::Split => cmd::split::run(argv),
             Command::Stats => cmd::stats::run(argv),
             Command::Table => cmd::table::run(argv),
             Command::Transpose => cmd::transpose::run(argv),
             Command::Validate => cmd::validate::run(argv),
-            #[cfg(feature = "foreach")]
-            Command::ForEach => cmd::foreach::run(argv),
         }
     }
 }
