@@ -5,7 +5,7 @@ use std::path::Path;
 
 #[test]
 fn generate_schema_with_defaults_and_validate_with_no_errors() {
-    // create worksapce and invoke schema command with value constraints flag
+    // create workspace and invoke schema command with value constraints flag
     let wrk = Workdir::new("schema").flexible(true);
     wrk.clear_contents().unwrap();
 
@@ -32,7 +32,7 @@ fn generate_schema_with_defaults_and_validate_with_no_errors() {
     // diff output json with expected json
     let expected_schema: String =
         wrk.load_test_resource("adur-public-toilets.csv.schema-default.expected.json");
-    let expected_schema_json: Value = serde_json::from_str(&expected_schema.to_string()).unwrap();
+    let expected_schema_json: Value = serde_json::from_str(&expected_schema).unwrap();
     assert_json_eq!(expected_schema_json, output_schema_json);
 
     // invoke validate command from schema created above
@@ -44,14 +44,14 @@ fn generate_schema_with_defaults_and_validate_with_no_errors() {
     // not expecting any invalid rows, so confirm there are NO output files generated
     let validation_error_path = &wrk.path("adur-public-toilets.csv.validation-errors.tsv");
     println!("not expecting validation error file at: {validation_error_path:?}");
-    assert!(Path::new(validation_error_path).exists() == false);
-    assert!(Path::new(&wrk.path("adur-public-toilets.csv.valid")).exists() == false);
-    assert!(Path::new(&wrk.path("adur-public-toilets.csv.invalid")).exists() == false);
+    assert!(!Path::new(validation_error_path).exists());
+    assert!(!Path::new(&wrk.path("adur-public-toilets.csv.valid")).exists());
+    assert!(!Path::new(&wrk.path("adur-public-toilets.csv.invalid")).exists());
 }
 
 #[test]
 fn generate_schema_with_optional_flags_and_validate_with_errors() {
-    // create worksapce and invoke schema command with value constraints flag
+    // create workspace and invoke schema command with value constraints flag
     let wrk = Workdir::new("schema").flexible(true);
     wrk.clear_contents().unwrap();
 
@@ -83,7 +83,7 @@ fn generate_schema_with_optional_flags_and_validate_with_errors() {
     // diff output json with expected json
     let expected_schema: String =
         wrk.load_test_resource("adur-public-toilets.csv.schema-strict.expected.json");
-    let expected_schema_json: Value = serde_json::from_str(&expected_schema.to_string()).unwrap();
+    let expected_schema_json: Value = serde_json::from_str(&expected_schema).unwrap();
     assert_json_eq!(expected_schema_json, output_schema_json);
 
     // invoke validate command from schema created above
@@ -114,18 +114,42 @@ fn generate_schema_with_optional_flags_and_validate_with_errors() {
     let validation_error_path = &wrk.path("adur-public-toilets.csv.validation-errors.tsv");
     println!("expecting validation error file at: {validation_error_path:?}");
 
-    assert!(Path::new(validation_error_path).exists() == true);
-    assert!(Path::new(&wrk.path("adur-public-toilets.csv.valid")).exists() == true);
-    assert!(Path::new(&wrk.path("adur-public-toilets.csv.invalid")).exists() == true);
+    assert!(Path::new(validation_error_path).exists());
+    assert!(Path::new(&wrk.path("adur-public-toilets.csv.valid")).exists());
+    assert!(Path::new(&wrk.path("adur-public-toilets.csv.invalid")).exists());
 
     // check validation error output
     let validation_error_output: String =
         wrk.from_str(&wrk.path("adur-public-toilets.csv.validation-errors.tsv"));
 
-    assert!(validation_error_output.len() > 0);
+    assert!(!validation_error_output.is_empty());
 
     assert_eq!(
         validation_errors_expected.to_string(),
         validation_error_output
     );
 }
+
+#[test]
+fn generate_schema_with_defaults_to_stdout() {
+    // create workspace and invoke schema command with value constraints flag
+    let wrk = Workdir::new("schema_stdout").flexible(true);
+    wrk.clear_contents().unwrap();
+
+    // copy csv file to workdir
+    let csv = wrk.load_test_resource("adur-public-toilets.csv");
+    wrk.create_from_string("adur-public-toilets.csv", &csv);
+
+    // run schema command
+    let mut cmd = wrk.command("schema");
+    cmd.arg("--stdout").arg("adur-public-toilets.csv");
+    let stdout = wrk.stdout::<String>(&mut cmd);
+    let output_schema_json: Value = serde_json::from_str(&stdout).unwrap();
+
+    // diff output json with expected json
+    let expected_schema: String =
+        wrk.load_test_resource("adur-public-toilets.csv.schema-default.expected.json");
+    let expected_schema_json: Value = serde_json::from_str(&expected_schema).unwrap();
+
+    assert_json_eq!(expected_schema_json, output_schema_json);
+}
