@@ -8,26 +8,26 @@ macro_rules! stats_tests {
     ($name:ident, $field:expr, $rows:expr, $expect:expr) => {
         stats_tests!($name, $field, $rows, $expect, false, true);
     };
-    ($name:ident, $field:expr, $rows:expr, $expect:expr, $nulls:expr, $dates:expr) => {
+    ($name:ident, $field:expr, $rows:expr, $expect:expr, $nulls:expr, $infer_dates:expr) => {
         mod $name {
             use super::test_stats;
 
-            stats_test_headers!($name, $field, $rows, $expect, $nulls, $dates);
-            stats_test_no_headers!($name, $field, $rows, $expect, $nulls, $dates);
+            stats_test_headers!($name, $field, $rows, $expect, $nulls, $infer_dates);
+            stats_test_no_headers!($name, $field, $rows, $expect, $nulls, $infer_dates);
         }
     };
 }
 
-macro_rules! stats_nodates_tests {
+macro_rules! stats_no_infer_dates_tests {
     ($name:ident, $field:expr, $rows:expr, $expect:expr) => {
         stats_tests!($name, $field, $rows, $expect, false, false);
     };
-    ($name:ident, $field:expr, $rows:expr, $expect:expr, $nulls:expr, $dates:expr) => {
+    ($name:ident, $field:expr, $rows:expr, $expect:expr, $nulls:expr, $infer_dates:expr) => {
         mod $name {
             use super::test_stats;
 
-            stats_test_headers!($name, $field, $rows, $expect, $nulls, $dates);
-            stats_test_no_headers!($name, $field, $rows, $expect, $nulls, $dates);
+            stats_test_headers!($name, $field, $rows, $expect, $nulls, $infer_dates);
+            stats_test_no_headers!($name, $field, $rows, $expect, $nulls, $infer_dates);
         }
     };
 }
@@ -36,17 +36,35 @@ macro_rules! stats_test_headers {
     ($name:ident, $field:expr, $rows:expr, $expect:expr) => {
         stats_test_headers!($name, $field, $rows, $expect, false, true);
     };
-    ($name:ident, $field:expr, $rows:expr, $expect:expr, $nulls:expr, $dates:expr) => {
+    ($name:ident, $field:expr, $rows:expr, $expect:expr, $nulls:expr, $infer_dates:expr) => {
         #[test]
         fn headers_no_index() {
             let name = concat!(stringify!($name), "_headers_no_index");
-            test_stats(name, $field, $rows, $expect, true, false, $nulls, $dates);
+            test_stats(
+                name,
+                $field,
+                $rows,
+                $expect,
+                true,
+                false,
+                $nulls,
+                $infer_dates,
+            );
         }
 
         #[test]
         fn headers_index() {
             let name = concat!(stringify!($name), "_headers_index");
-            test_stats(name, $field, $rows, $expect, true, true, $nulls, $dates);
+            test_stats(
+                name,
+                $field,
+                $rows,
+                $expect,
+                true,
+                true,
+                $nulls,
+                $infer_dates,
+            );
         }
     };
 }
@@ -55,21 +73,40 @@ macro_rules! stats_test_no_headers {
     ($name:ident, $field:expr, $rows:expr, $expect:expr) => {
         stats_test_no_headers!($name, $field, $rows, $expect, false, true);
     };
-    ($name:ident, $field:expr, $rows:expr, $expect:expr, $nulls:expr, $dates:expr) => {
+    ($name:ident, $field:expr, $rows:expr, $expect:expr, $nulls:expr, $infer_dates:expr) => {
         #[test]
         fn no_headers_no_index() {
             let name = concat!(stringify!($name), "_no_headers_no_index");
-            test_stats(name, $field, $rows, $expect, false, false, $nulls, $dates);
+            test_stats(
+                name,
+                $field,
+                $rows,
+                $expect,
+                false,
+                false,
+                $nulls,
+                $infer_dates,
+            );
         }
 
         #[test]
         fn no_headers_index() {
             let name = concat!(stringify!($name), "_no_headers_index");
-            test_stats(name, $field, $rows, $expect, false, true, $nulls, $dates);
+            test_stats(
+                name,
+                $field,
+                $rows,
+                $expect,
+                false,
+                true,
+                $nulls,
+                $infer_dates,
+            );
         }
     };
 }
 
+#[allow(clippy::too_many_arguments)]
 fn test_stats<S>(
     name: S,
     field: &str,
@@ -78,15 +115,21 @@ fn test_stats<S>(
     headers: bool,
     use_index: bool,
     nulls: bool,
-    dates: bool,
+    infer_dates: bool,
 ) where
     S: ::std::ops::Deref<Target = str>,
 {
-    let (wrk, mut cmd) = setup(name, rows, headers, use_index, nulls, dates);
+    let (wrk, mut cmd) = setup(name, rows, headers, use_index, nulls, infer_dates);
     let field_val = get_field_value(&wrk, &mut cmd, field);
     // Only compare the first few bytes since floating point arithmetic
     // can mess with exact comparisons.
-    let len = cmp::min(11, cmp::min(field_val.len(), expected.len()));
+    // when field = skewness, we're comparing a long sequence of the quartile columns,
+    // that's why we use 40, if not, its a single column, and we need to compare only
+    // the first 15 characters just in case its a float
+    let len = cmp::min(
+        if field == "skewness" { 40 } else { 15 },
+        cmp::min(field_val.len(), expected.len()),
+    );
     assert_eq!(&field_val[0..len], &expected[0..len]);
 }
 
@@ -96,7 +139,7 @@ fn setup<S>(
     headers: bool,
     use_index: bool,
     nulls: bool,
-    dates: bool,
+    infer_dates: bool,
 ) -> (Workdir, process::Command)
 where
     S: ::std::ops::Deref<Target = str>,
@@ -120,17 +163,14 @@ where
     if nulls {
         cmd.arg("--nulls");
     }
-    if dates {
-        cmd.arg("--dates");
+    if infer_dates {
+        cmd.arg("--infer-dates").arg("--dates-whitelist").arg("all");
     }
 
     (wrk, cmd)
 }
 
 fn get_field_value(wrk: &Workdir, cmd: &mut process::Command, field: &str) -> String {
-    if field == "nullcount" {
-        cmd.arg("--nullcount");
-    }
     if field == "median" {
         cmd.arg("--median");
     }
@@ -143,8 +183,8 @@ fn get_field_value(wrk: &Workdir, cmd: &mut process::Command, field: &str) -> St
     if field == "mode" {
         cmd.arg("--mode");
     }
-    if field == "dates" {
-        cmd.arg("--dates");
+    if field == "infer_dates" {
+        cmd.arg("--infer-dates");
     }
 
     let mut rows: Vec<Vec<String>> = wrk.read_stdout(cmd);
@@ -154,12 +194,13 @@ fn get_field_value(wrk: &Workdir, cmd: &mut process::Command, field: &str) -> St
         for (h, val) in headers.iter().zip(row.iter()) {
             match field {
                 "quartiles" => match &**h {
-                    "lower_fence" | "q1" | "q2_median" | "q3" | "iqr" => {
+                    "lower_outer_fence" | "lower_inner_fence" | "q1" | "q2_median" | "q3"
+                    | "iqr" | "upper_inner_fence" | "upper_outer_fence" => {
                         sequence.push(val);
                     }
-                    "upper_fence" => {
+                    "skewness" => {
                         sequence.push(val);
-                        return sequence.join(",").clone();
+                        return sequence.join(",");
                     }
                     _ => {}
                 },
@@ -183,14 +224,14 @@ stats_tests!(stats_infer_int, "type", &["1"], "Integer");
 stats_tests!(stats_infer_float, "type", &["1.2"], "Float");
 stats_tests!(stats_infer_null, "type", &[""], "NULL");
 stats_tests!(stats_infer_date, "type", &["1968-06-27"], "Date");
-stats_nodates_tests!(stats_infer_nodate, "type", &["1968-06-27"], "String");
+stats_no_infer_dates_tests!(stats_infer_nodate, "type", &["1968-06-27"], "String");
 stats_tests!(
     stats_infer_datetime,
     "type",
     &["1968-06-27 12:30:01"],
     "DateTime"
 );
-stats_nodates_tests!(
+stats_no_infer_dates_tests!(
     stats_infer_nodatetime,
     "type",
     &["1968-06-27 12:30:01"],
@@ -205,8 +246,8 @@ stats_tests!(
     &["June 27, 1968", ""],
     "Date"
 );
-stats_nodates_tests!(
-    stats_infer_nodates_null,
+stats_no_infer_dates_tests!(
+    stats_infer_no_infer_dates_null,
     "type",
     &["June 27, 1968", ""],
     "String"
@@ -217,7 +258,7 @@ stats_tests!(
     &["June 27, 1968 12:30:00 UTC", ""],
     "DateTime"
 );
-stats_nodates_tests!(
+stats_no_infer_dates_tests!(
     stats_infer_nodatetime_null,
     "type",
     &["June 27, 1968 12:30:00 UTC", ""],
@@ -232,7 +273,7 @@ stats_tests!(
     &["", "September 17, 2012 at 10:09am PST"],
     "Date"
 );
-stats_nodates_tests!(
+stats_no_infer_dates_tests!(
     stats_infer_null_nodate,
     "type",
     &["", "September 17, 2012 at 10:09am PST"],
@@ -244,7 +285,7 @@ stats_tests!(
     &["September 11, 2001", "September 17, 2012 at 10:09am PST"],
     "DateTime"
 );
-stats_nodates_tests!(
+stats_no_infer_dates_tests!(
     stats_infer_nodate_nodatetime,
     "type",
     &["September 11, 2001", "September 17, 2012 at 10:09am PST"],
@@ -357,6 +398,19 @@ stats_tests!(stats_sum_mixed2, "sum", &["2", "1.5"], "3.5");
 stats_tests!(stats_sum_mixed3, "sum", &["1.5", "hi", "2.8"], "4.3");
 stats_tests!(stats_sum_nulls1, "sum", &["1", "", "2"], "3");
 stats_tests!(stats_sum_nulls2, "sum", &["", "1", "2"], "3");
+stats_tests!(
+    stats_sum_overflow,
+    "sum",
+    &[
+        {
+            let i = i64::MAX;
+            &i.to_string()
+        },
+        "1",
+        "2"
+    ],
+    "OVERFLOW"
+);
 
 stats_tests!(stats_min, "min", &["2", "1.1"], "1.1");
 stats_tests!(stats_max, "max", &["2", "1.1"], "2");
@@ -422,32 +476,33 @@ stats_tests!(
     stats_quartiles,
     "quartiles",
     &["1", "2", "3"],
-    "-2,1,2,3,2,4"
+    "-5.0,-2.0,1.0,2.0,3.0,2.0,6.0,9.0"
 );
 stats_tests!(
     stats_quartiles_null,
     "quartiles",
     &["", "1", "2", "3"],
-    "-2,1,2,3,2,4"
+    "-5.0,-2.0,1.0,2.0,3.0,2.0,6.0,9.0"
 );
 stats_tests!(
     stats_quartiles_even,
     "quartiles",
     &["1", "2", "3", "4"],
-    "-1.5,1.5,2.5,3.5,1,4.5"
+    "-4.5,-1.5,1.5,2.5,3.5,2.0,6.5,9.5"
 );
 stats_tests!(
     stats_quartiles_even_null,
     "quartiles",
     &["", "1", "2", "3", "4"],
-    "-1.5,1.5,2.5,3.5,1,4.5"
+    "-4.5,-1.5,1.5,2.5,3.5,2.0,6.5,9.5"
 );
 stats_tests!(
     stats_quartiles_mix,
     "quartiles",
     &["1", "2.0", "3", "4"],
-    "-1.5,1.5,2.5,3.5,1,4.5"
+    "-4.5,-1.5,1.5,2.5,3.5,2.0,6.5,9.5"
 );
+stats_tests!(stats_quartiles_null_empty, "quartiles", &[""], "");
 
 stats_tests!(stats_nullcount, "nullcount", &["", "1", "2"], "1");
 stats_tests!(stats_nullcount_none, "nullcount", &["a", "1", "2"], "0");
@@ -458,6 +513,109 @@ stats_tests!(
     "0"
 );
 stats_tests!(stats_nullcount_all, "nullcount", &["", "", ""], "3");
+stats_no_infer_dates_tests!(
+    stats_noinfer_null_nodate2,
+    "type",
+    &["", "September 17, 2012 at 10:09am PST"],
+    "String",
+    true,
+    false
+);
+stats_no_infer_dates_tests!(
+    stats_infer_null_nodate2,
+    "type",
+    &["", "September 17, 2012 at 10:09am PST"],
+    "DateTime",
+    true,
+    true
+);
+stats_tests!(
+    stats_infer_date_datetime2,
+    "type",
+    &["September 11, 2001", "September 17, 2012 at 10:09am PST"],
+    "DateTime",
+    false,
+    true
+);
+stats_no_infer_dates_tests!(
+    stats_infer_date_datetime3,
+    "type",
+    &["9-11", "September 17, 2012 at 10:09am PST"],
+    "String",
+    false,
+    true
+);
+
+#[test]
+fn stats_prefer_dmy() {
+    let wrk = Workdir::new("stats_prefer_dmy");
+    let test_file = wrk.load_test_file("boston311-dmy-100.csv");
+
+    let mut cmd = wrk.command("stats");
+    cmd.arg("--infer-dates")
+        .arg("--prefer-dmy")
+        .arg("--dates-whitelist")
+        .arg("_dt")
+        .arg(test_file);
+
+    let got: String = wrk.stdout(&mut cmd);
+
+    let expected = wrk.load_test_resource("boston311-100-stats.csv");
+
+    assert_eq!(got, expected.replace("\r\n", "\n").trim_end());
+}
+
+#[test]
+fn stats_prefer_mdy() {
+    let wrk = Workdir::new("stats_prefer_mdy");
+    let test_file = wrk.load_test_file("boston311-100.csv");
+
+    let mut cmd = wrk.command("stats");
+    cmd.arg("--infer-dates")
+        .arg("--dates-whitelist")
+        .arg("_dt")
+        .arg(test_file);
+
+    let got: String = wrk.stdout(&mut cmd);
+
+    let expected = wrk.load_test_resource("boston311-100-stats.csv");
+
+    assert_eq!(got, expected.replace("\r\n", "\n").trim_end());
+}
+
+#[test]
+fn stats_no_date_inference() {
+    let wrk = Workdir::new("stats_no_date_inference");
+    let test_file = wrk.load_test_file("boston311-100.csv");
+
+    let mut cmd = wrk.command("stats");
+    cmd.arg("--everything").arg(test_file);
+
+    let got: String = wrk.stdout(&mut cmd);
+
+    let expected = wrk.load_test_resource("boston311-100-everything-nodate-stats.csv");
+
+    assert_eq!(got, expected.replace("\r\n", "\n").trim_end());
+}
+
+#[test]
+fn stats_with_date_inference() {
+    let wrk = Workdir::new("stats_with_date_inference");
+    let test_file = wrk.load_test_file("boston311-100.csv");
+
+    let mut cmd = wrk.command("stats");
+    cmd.arg("--everything")
+        .arg(test_file)
+        .arg("--infer-dates")
+        .arg("--dates-whitelist")
+        .arg("all");
+
+    let got: String = wrk.stdout(&mut cmd);
+
+    let expected = wrk.load_test_resource("boston311-100-everything-date-stats.csv");
+
+    assert_eq!(got, expected.replace("\r\n", "\n").trim_end());
+}
 
 mod stats_infer_nothing {
     // Only test CSV data with headers.
