@@ -8,6 +8,7 @@ extern crate quickcheck;
 extern crate rand;
 extern crate stats;
 
+use std::env;
 use std::fmt;
 use std::mem::transmute;
 use std::ops;
@@ -42,8 +43,11 @@ mod test_comments;
 mod test_count;
 mod test_dedup;
 mod test_enumerate;
+mod test_excel;
 mod test_exclude;
 mod test_explode;
+mod test_extsort;
+#[cfg(feature = "fetch")]
 mod test_fetch;
 mod test_fill;
 mod test_fixlengths;
@@ -71,6 +75,7 @@ mod test_search;
 mod test_searchset;
 mod test_select;
 mod test_slice;
+mod test_sniff;
 mod test_sort;
 mod test_split;
 mod test_stats;
@@ -79,11 +84,15 @@ mod test_transpose;
 mod test_validate;
 
 fn qcheck<T: Testable>(p: T) {
+    env::set_var("QSV_SKIPUTF8_CHECK", "1");
     QuickCheck::new().gen(Gen::new(5)).quickcheck(p);
+    env::set_var("QSV_SKIPUTF8_CHECK", "");
 }
 
 fn qcheck_sized<T: Testable>(p: T, size: usize) {
+    env::set_var("QSV_SKIPUTF8_CHECK", "1");
     QuickCheck::new().gen(Gen::new(size)).quickcheck(p);
+    env::set_var("QSV_SKIPUTF8_CHECK", "");
 }
 
 pub type CsvVecs = Vec<Vec<String>>;
@@ -112,6 +121,7 @@ impl CsvRecord {
     }
 }
 
+#[allow(clippy::needless_lifetimes)]
 impl ops::Deref for CsvRecord {
     type Target = [String];
     fn deref<'a>(&'a self) -> &'a [String] {
@@ -119,6 +129,7 @@ impl ops::Deref for CsvRecord {
     }
 }
 
+#[allow(clippy::needless_lifetimes)]
 impl ops::DerefMut for CsvRecord {
     fn deref_mut<'a>(&'a mut self) -> &'a mut [String] {
         &mut *self.0
@@ -177,6 +188,7 @@ impl CsvData {
     }
 }
 
+#[allow(clippy::needless_lifetimes)]
 impl ops::Deref for CsvData {
     type Target = [CsvRecord];
     fn deref<'a>(&'a self) -> &'a [CsvRecord] {
@@ -198,7 +210,7 @@ impl Arbitrary for CsvData {
         // If the CSV data starts with a BOM, strip it, because it wreaks havoc
         // with tests that weren't designed to handle it.
         if !d.data.is_empty() && !d.data[0].is_empty() {
-            if let Some(stripped) = d.data[0][0].strip_prefix("\u{FEFF}") {
+            if let Some(stripped) = d.data[0][0].strip_prefix('\u{FEFF}') {
                 d.data[0][0] = stripped.to_string();
             }
         }
